// SPDX-License-Identifier: GPL-2.0
/*
 * Sony ifrtp SFP Ethernet device driver
 *
 * Copyright (c) 2023 Sony Corporation. All rights reserved.
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/moduleparam.h>

#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/workqueue.h>

#include <linux/in.h>
#include <linux/etherdevice.h>
#include <linux/ip.h>
#include <linux/if_arp.h>
#include <linux/tcp.h>
#include <linux/skbuff.h>
#include <linux/version.h>
#include <linux/dma-direction.h>
#include <linux/dma-mapping.h>
#include <linux/dma-map-ops.h>
#include <linux/dmaengine.h>
#include <linux/poison.h>
#include <linux/string.h>
#include <linux/gfp.h>

#include "i2c.h"
#include "ifrtp.h"
#include "ifrtp_address.h"
#include "ifrtp_logging.h"
#include "ifrtp_driver.h"
#include "ifrtp_filter.h"
#include "ifrtp_ioctl.h"
#include "ifrtp_mdio.h"
#include "io.h"
#include "igw_driver.h"
#include "igw_netreg.h"
#include "irq.h"
#include <linux/if_link.h>
#include <linux/socket.h>
#include <linux/hrtimer.h>
#include <linux/time.h>
#include <linux/ktime.h>
#include <linux/ethtool.h>
#include <linux/inetdevice.h>
#include <linux/notifier.h>
#include <linux/irqreturn.h>
#include <linux/netdevice.h>

#include <linux/in6.h>
#include <asm/checksum.h>
#include <linux/processor.h>


#define IFRTP_TX_DESCRIPTOR_DEPTH	(ifrtp_cfg.tx_buffer)
#define IFRTP_RX_DESCRIPTOR_DEPTH	(ifrtp_cfg.rx_buffer)

#define IFRTP_DMA_CHANNEL_0		""
#define IFRTP_DMA_CHANNEL_1		""

#define ESTIMATED_BOOTTIME_SECONDS	64

DEFINE_SPINLOCK(tx_desc_lock);
DEFINE_SPINLOCK(rx_desc_lock);
struct sk_buff **tx_skbuff_p;
u32 *desc_addr_tx;
u32 *desc_addr_rx, *packet_addr_rx;
u64 desc_addr_tx_phys, desc_addr_rx_phys, packet_addr_rx_phys;
#ifdef IFRTP_RX_USE_DMAC_MEMCPY
	struct dma_chan *dmachan_memcpy;
#endif
struct ifrtp_recv_mem g_recv_buff[RTP_SSRC_MAX];

struct napi_struct ifrtp_napi;

#ifdef IFRTP_ENABLE_DEBUGFS
	struct dentry *ifrtp_debugfs_dir;
	struct debugfs_u32_array debugfs_arrary_1;
	u32 debugfs_counter_1;
	u32 debugfs_var_1;
	u32 debugfs_trigger_mdio_core_reset = 0;
	u32 debugfs_i2c_module_0;
	u32 debugfs_i2c_module_1;
	u32 debugfs_mdio_0;
	u32 debugfs_mdio_1;
	bool debugfs_link_no_autoreset = false;
	bool debugfs_link_sw_force_up = false;
#endif

static int ifrtp_release(struct net_device *dev)
{
	// release ports, irq
	netif_stop_queue(dev);
	return 0;
}

static int ifrtp_config(struct net_device *dev, struct ifmap *map)
{
	if (dev->flags & IFF_UP)
		return -EBUSY;

	if (map->base_addr != dev->base_addr) {
		pr_warn("Can't change I/O address\n");
		return -EOPNOTSUPP;
	}

	if (map->irq != dev->irq) {
		pr_warn("Can't change IRQ\n");
		return -EOPNOTSUPP;
	}

	return 0;
}

struct net_device *ifrtp_devs[IFRTP_NUMVFS_MAX];

static int ifrtp_open(struct net_device *dev)
{
	netif_start_queue(dev);

	return 0;
}

irqreturn_t ifrtp_napi_schedule(void)
{
	if (ifrtp_get_net_rx_irq_status() && napi_schedule_prep(&ifrtp_napi)) {
		ifrtp_disable_net_rx_irq();
		__napi_schedule_irqoff(&ifrtp_napi);
		return IRQ_HANDLED;
	}
	return IRQ_NONE;
}


int ifrtp_send_direct_multi_message(unsigned char *data, unsigned int len, unsigned int packet_type, struct sk_buff *skb)
{
	static unsigned int wp;
	static unsigned int prev_rp;
	int i, rp = 0, batch_write_count = 0, res = 0;
	u32 transmitted = 0, vlen = len / sizeof(struct ifrtp_msghdr);
	struct ifrtp_msghdr *msg_vec = (struct ifrtp_msghdr *)data;

	spin_lock(&tx_desc_lock);

	// no need for something like
	// wp = ifrtp_ior32(NULL, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_WP);
	// use static variable to reduce ioread count
	rp = ifrtp_ior32(NULL, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_RP);

	for (i = 0; i * sizeof(struct ifrtp_msghdr) < len; ++i) {
		if (unlikely(rp == ((wp+1) & (ifrtp_cfg.tx_buffer-1)))) {
			rp = ifrtp_ior32(NULL, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_RP);
			if (unlikely(rp == ((wp+1) & (ifrtp_cfg.tx_buffer-1)))) {
				// Buffer full
				transmitted = i;
				res = -EAGAIN;
				goto trans_err;
			}
		}

		// sync for raw packet
		if (packet_type & (1 << 16)) {
			msg_vec[i].msg_addr = dma_map_single(&ifrtp_cfg.pdev->dev,
					(void *)msg_vec[i].msg_addr,
					msg_vec[i].msg_len,
					DMA_TO_DEVICE);

			if (dma_mapping_error(&ifrtp_cfg.pdev->dev, msg_vec[i].msg_addr)) {
				transmitted = i;
				res = -ENOMEM;
				kfree_skb(skb);
				goto trans_err;
			}
		}

		desc_addr_tx[wp*4]	= msg_vec[i].msg_addr >> 32;
		desc_addr_tx[wp*4+1]	= msg_vec[i].msg_addr & 0xFFFFFFFF;
		desc_addr_tx[wp*4+2]	= msg_vec[i].msg_len;
		desc_addr_tx[wp*4+3]	= packet_type;

		tx_skbuff_p[wp] = skb;

		wp = (wp+1) & (ifrtp_cfg.tx_buffer-1);
		++batch_write_count;
		if (batch_write_count >= ifrtp_cfg.tx_write_batch || wp == 0) {
			dma_sync_single_for_device(&ifrtp_cfg.pdev->dev,
					(dma_addr_t)desc_addr_tx_phys + ((wp == 0 ? ifrtp_cfg.tx_buffer : wp) - batch_write_count) * 16,
					batch_write_count * 16,
					DMA_TO_DEVICE);
			ifrtp_iow32(NULL, wp, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_WP);
			batch_write_count = 0;
		}
	}

	transmitted = vlen;
trans_err:
	for (; prev_rp != rp; prev_rp = (prev_rp + 1) & (ifrtp_cfg.tx_buffer - 1)) {
		if (desc_addr_tx[prev_rp * 4 + 3] & (1 << 16)) {
			dma_unmap_single(&ifrtp_cfg.pdev->dev,
					((dma_addr_t)desc_addr_tx[prev_rp * 4] << 32) + desc_addr_tx[prev_rp * 4 + 1],
					desc_addr_tx[prev_rp * 4 + 2],
					DMA_TO_DEVICE);
			consume_skb(tx_skbuff_p[prev_rp]);
			tx_skbuff_p[prev_rp] = NULL;
		}
	}

	if (batch_write_count) {
		dma_sync_single_for_device(&ifrtp_cfg.pdev->dev,
				(dma_addr_t)desc_addr_tx_phys + (wp - batch_write_count) * 16,
				batch_write_count * 16,
				DMA_TO_DEVICE);
		ifrtp_iow32(NULL, wp, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_WP);
	}
	spin_unlock(&tx_desc_lock);

	return res;
}

static netdev_tx_t ifrtp_tx(struct sk_buff *skb, struct net_device *dev)
{
	struct ifrtp_priv *priv = netdev_priv(dev);

	if (unlikely(ntohs(skb->protocol) != 0)) {
		kfree_skb(skb);
		return NETDEV_TX_OK;
	}

	if (unlikely(ifrtp_send_direct_multi_message(skb->data, skb->len, priv->stream_id, NULL) < 0)) {
		netdev_err_once(dev, "descriptor buffer full\n");
		priv->stats.tx_errors++;
		priv->stats.tx_fifo_errors++;
		// sk_buff must NOT be freed if NETDEV_TX_BUSY is returned
		// TODO (guo): properly control netif queue stop/wake
		// return NETDEV_TX_BUSY;
	}

	// No need to check if skb == NULL, consume_skb will do it.
	// dev_consume_skb_irq(skb);
	consume_skb(skb);
	return NETDEV_TX_OK;
}


static void ifrtp_tx_timeout(struct net_device *dev, unsigned int txqueue)
{
	struct ifrtp_priv *priv = netdev_priv(dev);
	struct netdev_queue *txq = netdev_get_tx_queue(dev, 0);

	ifrtp_pr_debug("Transmit timeout at %ld, latency %ld\n", jiffies,
			jiffies - txq->trans_start);
	// TODO (guo) control transmission interrupt?
	priv->stats.tx_errors++;

	netif_wake_queue(dev);
}


static void ifrtp_stats64(struct net_device *dev, struct rtnl_link_stats64 *storage)
{
	struct ifrtp_priv *priv = netdev_priv(dev);
	*storage = priv->stats;
}


// TODO (guo) check if necessery
static int ifrtp_change_mtu(struct net_device *dev, int new_mtu)
{
	if ((new_mtu < dev->min_mtu) || (new_mtu > dev->max_mtu))
		return -EINVAL;

	dev->mtu = new_mtu;
	return 0;
}


static int ifrtp_siocdevprivate(struct net_device *dev, struct ifreq *ifr, void __user *data, int cmd)
{
	return ifrtp_ioctl(dev, ifr, cmd);
}

static const struct net_device_ops ifrtp_netdev_ops = {
	.ndo_open		= ifrtp_open,
	.ndo_stop		= ifrtp_release,
	.ndo_start_xmit		= ifrtp_tx,
#if KERNEL_VERSION(5, 15, 0) <= LINUX_VERSION_CODE
	.ndo_siocdevprivate	= ifrtp_siocdevprivate,
#else
	.ndo_do_ioctl		= ifrtp_ioctl,
#endif
	.ndo_set_config		= ifrtp_config,
	.ndo_get_stats64	= ifrtp_stats64,
	.ndo_change_mtu		= ifrtp_change_mtu,
	.ndo_tx_timeout		= ifrtp_tx_timeout,
};

void __iomem *rtp_recv_buffer_remaped, *rtp_recv_buffer_remap_offset;

static inline enum ifrtp_interface_type network_type_to_interface_type(unsigned int type)
{
	if (type < IFRTP_SDI_CH_MAX_PER_CORE)
		return IFRTP_IF_IVFRTP;

	if (type == 0x1E || type == 0x1F)
		return IFRTP_IF_IGW;

	return IFRTP_IF_ERROR;
}

static int ifrtp_poll(struct napi_struct *napi, int budget)
{
	int npackets = 0;

	// Regarding the order of reading rp and wp:
	// this is safe since only one thread is polling
	// read wp later to process more packet
	static unsigned int rp;
	// unsigned int rp = ifrtp_ior32(NULL, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_RP);
	unsigned int wp = ifrtp_ior32(NULL, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_WP);
	unsigned int cleaned_count = 0;

	while (wp != rp && npackets < budget) {
		dma_sync_single_for_cpu(&ifrtp_cfg.pdev->dev,
				(dma_addr_t)desc_addr_rx_phys,
				IFRTP_RX_DESCRIPTOR_DEPTH * 16,
				DMA_FROM_DEVICE);

		// TODO (guo): reduce sync size?
		if (wp > rp) {
			dma_sync_single_for_cpu(&ifrtp_cfg.pdev->dev,
					(dma_addr_t)packet_addr_rx_phys + rp * 2048,
					(wp - rp) * 2048,
					DMA_FROM_DEVICE);
		} else {
			dma_sync_single_for_cpu(&ifrtp_cfg.pdev->dev,
					(dma_addr_t)packet_addr_rx_phys,
					wp * 2048,
					DMA_FROM_DEVICE);
			dma_sync_single_for_cpu(&ifrtp_cfg.pdev->dev,
					(dma_addr_t)packet_addr_rx_phys + rp * 2048,
					(IFRTP_RX_DESCRIPTOR_DEPTH - rp) * 2048,
					DMA_FROM_DEVICE);
		}

		while (wp != rp && npackets < budget) {
			struct sk_buff *skb;
			struct ifrtp_priv *priv;
			unsigned int type	= desc_addr_rx[4*rp + 3] & 0xFF;
			unsigned int media_type	= (desc_addr_rx[4*rp + 3] >> 8 & 0xFF);
			enum ifrtp_interface_type if_type = network_type_to_interface_type(type);
			struct ifrtp_msghdr msg	= {
				.msg_len	= desc_addr_rx[4*rp + 2],
				.ssrc		= ntohl(*(__be32 *)(packet_addr_rx + 512*rp + 6)),
				.pt		= (ntohl(*(__be32 *)(packet_addr_rx + 512*rp + 4)) >> 16) & 0x7F,
				.if_no		= 0,
			};

			if (unlikely(if_type == IFRTP_IF_ERROR)) {
				pr_warn_ratelimited("invalid rx Network Type %d, packet dropped\n", type);
				goto rx_checkout;
			}

			if (if_type == IFRTP_IF_IGW) {
				struct igw_priv *priv_igw = ifrtp_cfg.driver_handle->dev[type - 0x1E];

				if (unlikely(msg.msg_len > ETH_FRAME_LEN || msg.msg_len < ETH_ZLEN)) {
					pr_warn_ratelimited("invalid packet length %zu, packet dropped\n", msg.msg_len);
					priv_igw->__netarch.stats.rx_length_errors++;
					priv_igw->__netarch.stats.rx_errors++;
					goto rx_checkout;
				}

				skb = __napi_alloc_skb(napi, msg.msg_len, GFP_ATOMIC | __GFP_NOWARN);
				if (unlikely(!skb)) {
					pr_warn_ratelimited("can't alloc sk_buff, packet dropped\n");
					priv_igw->__netarch.stats.rx_over_errors++;
					priv_igw->__netarch.stats.rx_errors++;
					goto rx_checkout;
				}

				memcpy(skb_put(skb, msg.msg_len), (unsigned char *)packet_addr_rx + 2048*rp + 16, msg.msg_len);
				skb->dev = priv_igw->__netarch.dev;
				skb->protocol = eth_type_trans(skb, skb->dev);
				skb->ip_summed = CHECKSUM_UNNECESSARY; // chucksum is done by FPGA

				// no need to check return value, netif_receive_skb will always success
				netif_receive_skb(skb);

				priv_igw->__netarch.stats.rx_packets++;
				priv_igw->__netarch.stats.rx_bytes += msg.msg_len;
				goto rx_checkout;
			}

			// ivfrtp packet
			priv = netdev_priv(ifrtp_devs[type*2]);

			if (unlikely(msg.msg_len > RTP_MAX_PACKET_SIZE || msg.msg_len < RTP_MIN_PACKET_SIZE)) {
				pr_warn_ratelimited("invalid packet length %zu, packet dropped\n", msg.msg_len);
				priv->stats.rx_length_errors++;
				priv->stats.rx_errors++;
				goto rx_checkout;
			}

			// SSRC mismatch
			if (unlikely(media_type >= RTP_SSRC_MAX || media_type < type * RTP_SSRC_PER_CH)) {
				struct ifrtp_priv *pp;

				if (media_type & 1) {
					pp = netdev_priv(ifrtp_devs[type*2+1]);
					msg.if_no = 1;
				} else {
					pp = priv;
				}

				pp->stats.rx_nohandler++;
				msg.msg_addr = 0;
				msg.msg_len = 0;
				ifrtp_pr_debug_ratelimited("potentially wrong ssrc 0x%08x from ch%d, recognized as media type %d\n",
						msg.ssrc, type, media_type);
				if (jiffies - pp->ssrc_mismatch_last < ifrtp_cfg.ssrc_mismatch_report_interval) {
					goto rx_checkout;
				} else {
					pp->ssrc_mismatch_last = jiffies;
					ifrtp_pr_debug_ratelimited("reporting wrong ssrc 0x%08x from WAN%d, pt:%02x, recognized as network type %d, media type %d\n", msg.ssrc, msg.if_no + 1, msg.pt, type, media_type);
				}
			} else {
				msg.msg_addr = priv->recv_buff[media_type].addr;
			}

			skb = __napi_alloc_skb(napi, sizeof(struct ifrtp_msghdr), GFP_ATOMIC | __GFP_NOWARN);
			if (unlikely(!skb)) {
				pr_warn_ratelimited("can't alloc sk_buff, packet dropped\n");
				priv->stats.rx_over_errors++;
				priv->stats.rx_errors++;
				goto rx_checkout;
			}

			if (likely(msg.msg_len)) {
#ifdef IFRTP_RX_USE_DMAC_MEMCPY
				// perform memcpy with dma channel 7
				struct dma_async_tx_descriptor *desc = dmaengine_prep_dma_memcpy(dmachan_memcpy,
						msg.msg_addr,
						packet_addr_rx_phys + 2048*rp + 16,
						msg.msg_len, DMA_CTRL_ACK);
				if (!desc) {
					pr_err_ratelimited("Failed to prepare DMA transfer\n");
					priv->stats.rx_errors++;
					goto rx_checkout;
				}

				dma_cookie_t cookie = desc->tx_submit(desc);

				if (dma_submit_error(cookie)) {
					pr_err_ratelimited("Failed to submit DMA transfer\n");
					priv->stats.rx_errors++;
					goto rx_checkout;
				}

				// dma_async_issue_pending(dmachan_memcpy);
				// don't checkout ?

				enum dma_status status = dma_sync_wait(dmachan_memcpy, cookie);

				if (status != DMA_COMPLETE)
					pr_err_ratelimited("Failed to perform DMA transfer\n");
#else
				memcpy(msg.msg_addr + rtp_recv_buffer_remap_offset,
						(unsigned char *)packet_addr_rx + 2048*rp + 16, msg.msg_len);
#endif

				priv->recv_buff[media_type].addr += RTP_PACKET_OFFSET_SIZE;
				if (priv->recv_buff[media_type].addr >= priv->recv_buff[media_type].end)
					priv->recv_buff[media_type].addr = priv->recv_buff[media_type].begin;
			}

			memcpy(skb_put(skb, sizeof(struct ifrtp_msghdr)), &msg, sizeof(struct ifrtp_msghdr));

			skb->dev = ifrtp_devs[type*2];
			skb->protocol = 0;
			skb->ip_summed = CHECKSUM_UNNECESSARY; // chucksum is done by FPGA

			// no need to check return value, netif_receive_skb will always success
			netif_receive_skb(skb);

			// no need for something like
			// priv->stats.rx_packets++;
			// priv->stats.rx_bytes += msg.msg_len;
			// FPGA will do the job

rx_checkout:
			++npackets;
			++cleaned_count;
			rp = (rp+1) & (IFRTP_RX_DESCRIPTOR_DEPTH-1);
			if (cleaned_count >= ifrtp_cfg.rx_write_batch) {
				ifrtp_iow32(NULL, rp, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_RP);
				cleaned_count = 0;
			}
		}

		wp = ifrtp_ior32(NULL, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_WP);
	}

	if (cleaned_count)
		ifrtp_iow32(NULL, rp, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_RP);

	// processed all packets within budget
	if (npackets < budget) {
		if (napi_complete_done(napi, npackets))
			ifrtp_enable_net_rx_irq();
	}

	return npackets;
}

static struct workqueue_struct *arp_check_wq;
static void ifrtp_schedule_arp_check(struct work_struct *w);
static DECLARE_DELAYED_WORK(ifrtp_work_arp_check, ifrtp_schedule_arp_check);
/*
 * check arp cache periodicly
 */
static void ifrtp_schedule_arp_check(struct work_struct *w)
{
	int i;

	for (i = 0; i < ifrtp_cfg.numvfs; ++i) {
		if (!(ifrtp_devs[i]->flags & IFF_UP))
			continue;

		struct ifrtp_priv *priv = netdev_priv(ifrtp_devs[i]);
		struct IocNetTxDestMacAddressConfig mac_config = { .ch_no = priv->stream_id };
		struct IocNetTxDestIpAddressConfig ip_config = { .ch_no = priv->stream_id };
		unsigned char eth_dst[6];

		if (priv->is_main_mac) {
			ifrtp_get_tx_dest_primary_mac_address(NULL, &mac_config);
			ifrtp_get_tx_dest_primary_address(NULL, &ip_config);
		} else {
			ifrtp_get_tx_dest_secondary_mac_address(NULL, &mac_config);
			ifrtp_get_tx_dest_secondary_address(NULL, &ip_config);
		}

		if (arp_query(htonl(ip_config.ip_dst_addr),
					priv->is_main_mac ? 0 : 1,
					eth_dst)) {
			continue;
		}

		if (ip_config.ip_dst_addr)
			priv->stream_tx_ready = true;

		// update tx_dest_mac_address only if the addr has changed
		if (memcmp(eth_dst, mac_config.eth_dst, sizeof(eth_dst))) {
			memcpy(mac_config.eth_dst, eth_dst, sizeof(eth_dst));
			if (priv->is_main_mac)
				ifrtp_set_tx_dest_primary_mac_address(NULL, &mac_config);
			else
				ifrtp_set_tx_dest_secondary_mac_address(NULL, &mac_config);
		}
	}
	queue_delayed_work(arp_check_wq, &ifrtp_work_arp_check, HZ * IFRTP_ARP_SCHEDULE_INTERVAL);
}

static int ifrtp_get_coalesce(struct net_device *netdev,
				struct ethtool_coalesce *ec,
				struct kernel_ethtool_coalesce *kernel_coal,
				struct netlink_ext_ack *extack)
{
	return -ENOTSUPP;
}

static int ifrtp_set_coalesce(struct net_device *netdev,
				struct ethtool_coalesce *ec,
				struct kernel_ethtool_coalesce *kernel_coal,
				struct netlink_ext_ack *extack)
{
	return -ENOTSUPP;
}

static const struct ethtool_ops ifrtp_ethtool_ops = {
	.supported_coalesce_params	= ETHTOOL_COALESCE_RX_USECS,
	.get_coalesce			= ifrtp_get_coalesce,
	.set_coalesce			= ifrtp_set_coalesce,
};

void ifrtp_init_packet_counter(void)
{
	int i;
	// select type 0-7 and 0x1d (firewall)
	for (i = 0; i < ifrtp_cfg.numvfs / 2; ++i) {
		ifrtp_iow32(NULL, i, IFRTP_TX_PKT_COUNTER + IFRTP_OFFSET_PKT_COUNTER_SEL_TYPE + i * IFRTP_OFFSET_PKT_COUNTER_SEL_TYPE_OFFSET);
		ifrtp_iow32(NULL, i, IFRTP_RX_PKT_COUNTER + IFRTP_OFFSET_PKT_COUNTER_SEL_TYPE + i * IFRTP_OFFSET_PKT_COUNTER_SEL_TYPE_OFFSET);
		ifrtp_iow32(NULL, i, IFRTP_RX_PKT_COUNTER + IFRTP_RX_PKT_COUNTER_SUBOFFSET + IFRTP_OFFSET_PKT_COUNTER_SEL_TYPE + i * IFRTP_OFFSET_PKT_COUNTER_SEL_TYPE_OFFSET);
	}
	ifrtp_iow32(NULL, IFRTP_PKT_TYPE_FIREWALL, IFRTP_RX_PKT_COUNTER + IFRTP_OFFSET_PKT_COUNTER_SEL_TYPE + i * IFRTP_OFFSET_PKT_COUNTER_SEL_TYPE_OFFSET);
	ifrtp_iow32(NULL, IFRTP_PKT_TYPE_FIREWALL, IFRTP_RX_PKT_COUNTER + IFRTP_RX_PKT_COUNTER_SUBOFFSET + IFRTP_OFFSET_PKT_COUNTER_SEL_TYPE + i * IFRTP_OFFSET_PKT_COUNTER_SEL_TYPE_OFFSET);
}

unsigned int ifrtp_read_packet_counter(unsigned int counter_top, struct ifrtp_packet_count *pktcnt)
{
	unsigned int res, exist;

	ifrtp_iow32(NULL, 1, counter_top + IFRTP_OFFSET_PKT_COUNTER_TICK);

	res = read_poll_timeout(ifrtp_ior32, res, res == 1, 1, 50000, false,
			NULL, counter_top + IFRTP_OFFSET_PKT_COUNTER_TICK);
	exist = ifrtp_ior32(NULL, counter_top + IFRTP_OFFSET_PKT_COUNTER_EXIST) & IFRTP_PKT_COUNTER_EXIST_MASK;

	if (unlikely(res == -ETIMEDOUT))
		pr_warn("packet counter +0x%08x tick timedout, exist: 0x%08x\n", counter_top, exist);

	if (exist) {
		ifrtp_memcpy_fromio(NULL, (unsigned int *)pktcnt,
					counter_top + IFRTP_OFFSET_PKT_COUNTER_PKTCNT,
					sizeof(struct ifrtp_packet_count) / sizeof(unsigned int));
	}
	return exist;
}

void ifrtp_update_firewall_settings(void)
{
	static struct xt_table_info *last_private;
	struct xt_table *table;
	struct net_device *dev;
	struct net *net = NULL;

	dev = ifrtp_cfg.driver_handle->netdev[0];
	if (!dev || !(dev->flags & IFF_UP))
		return;

	net = dev_net(dev);
	table = xt_find_table_lock(net, AF_INET, "filter");
	if (IS_ERR(table)) {
		pr_err("error finding xt_table: %ld\n", PTR_ERR(table));
		return;
	}

	// append, delete, insert, etc. from/to iptables will change table->private
	// i.e. replace the old table with new one
	if (table->private != last_private) {
		ifrtp_pr_debug("New xt_table found. Add following rules to Identification module.\n");
		ifrtp_filter_apply_netfilter_udp_port(table->private);
		last_private = table->private;
	}

	xt_table_unlock(table);
}

static bool if_available(struct net_device *dev)
{
	return dev && (dev->flags & IFF_UP) && netif_carrier_ok(dev) && netif_running(dev);
}

static void ifrtp_update_packet_count(void)
{
	struct ifrtp_packet_count pktcnt;
	struct ifrtp_priv *priv;
	int i, j, exist;

	for (i = 0; i < IFRTP_PHY_COUNT; ++i) {
		exist = ifrtp_read_packet_counter(IFRTP_RX_PKT_COUNTER + IFRTP_RX_PKT_COUNTER_SUBOFFSET * i, &pktcnt);
		if (!exist)
			continue;

		for (j = 0; j < ifrtp_cfg.numvfs / 2; ++j) {
			if (exist & (1 << j)) {
				priv = netdev_priv(ifrtp_devs[j*2+i]);
				priv->stats.rx_packets += pktcnt.packet_count[j];
				priv->stats.rx_bytes += pktcnt.packet_size[j];
			}
		}
		if (exist & (1 << IFRTP_PKT_TYPE_FIREWALL))
			ifrtp_cfg.driver_handle->dev[i]->__netarch.stats.rx_nohandler += pktcnt.packet_count[j];
	}

	exist = ifrtp_read_packet_counter(IFRTP_TX_PKT_COUNTER, &pktcnt);
	if (!exist)
		return;

	for (i = 0; i < IFRTP_PHY_COUNT; ++i) {
		if (!if_available(ifrtp_cfg.driver_handle->netdev[i]))
			continue;

		for (j = 0; j < ifrtp_cfg.numvfs / 2; ++j) {
			if (!(exist & (1 << j)))
				continue;

			priv = netdev_priv(ifrtp_devs[j*2+i]);
			if (priv->stream_tx_ready) {
				priv->stats.tx_packets += pktcnt.packet_count[j];
				priv->stats.tx_bytes += pktcnt.packet_size[j];
			}
		}
	}
}

static void ifrtp_update_trm_bucket(void)
{
	static bool ifrtp_rx_trm_bucket_inited = true;
	struct net_device *dev0, *dev1;

	dev0 = ifrtp_cfg.driver_handle->netdev[0];
	dev1 = ifrtp_cfg.driver_handle->netdev[1];
	if (if_available(dev0) && if_available(dev1)) {
		if (!ifrtp_rx_trm_bucket_inited) {
			ifrtp_rx_trm_bucket_inited = true;
		}
	} else {
		ifrtp_rx_trm_bucket_inited = false;
	}
}

enum ifrtp_link_process {
	IFRTP_LINK_UP,
	IFRTP_LINK_NORMAL,
	IFRTP_LINK_DOWN,
};
static struct workqueue_struct *ifrtp_2hz_periodic_wq;
static void ifrtp_2hz_periodic_work(struct work_struct *w);
static DECLARE_DELAYED_WORK(ifrtp_work_2hz_periodic, ifrtp_2hz_periodic_work);
static void ifrtp_2hz_periodic_work(struct work_struct *w)
{
	int i, lane_status;
	static enum ifrtp_link_process link_stat[IFRTP_PHY_COUNT] = {};
	static bool mdio_reset_done[IFRTP_PHY_COUNT] = {};
	static int rx_datapath_reset_count[IFRTP_PHY_COUNT] = {};
	static bool enter_normal_process[IFRTP_PHY_COUNT] = {};
	static bool perform_pll_reset = true;

	queue_delayed_work(ifrtp_2hz_periodic_wq, &ifrtp_work_2hz_periodic, HZ / 2);

#ifdef IFRTP_ENABLE_DEBUGFS
	u16 mdio_low, mdio_high;

	__mdio_read16(0, &mdio_low, 0);
	__mdio_read16(1, &mdio_high, 0);
	debugfs_mdio_0 = ((u32)(mdio_high) << 16) + mdio_low;
	__mdio_read16(0, &mdio_low, 1);
	__mdio_read16(1, &mdio_high, 1);
	debugfs_mdio_1 = ((u32)(mdio_high) << 16) + mdio_low;
	debugfs_i2c_module_0 = ifrtp_i2c_module_read(0);
	debugfs_i2c_module_1 = ifrtp_i2c_module_read(1);

	for (i = 0; i < IFRTP_PHY_COUNT; ++i) {
		if (debugfs_trigger_mdio_core_reset & 1<<i) {
			debugfs_trigger_mdio_core_reset &= ~(1<<i);
			if (ifrtp_mdio_link_reset(i, ifrtp_cfg.mdio_reset_retry_interval_ms, ifrtp_cfg.mdio_reset_retry_times))
				pr_err("error reset pcs/pma through mdio %d\n", i);

		}
		if (debugfs_link_no_autoreset)
			goto exit_link_sequence;
	}
#endif

	// refre to https://support.xilinx.com/s/article/71820
	for (i = 0; i < IFRTP_PHY_COUNT; ++i) {
		lane_status = ifrtp_sfp_lane_status(i);
		ifrtp_pr_debug("WAN%d: lane_status = 0x%08x\n", i, lane_status);
		switch (link_stat[i]) {
		case IFRTP_LINK_UP:
			if (perform_pll_reset) {
				pr_warn("PCS/PMA Lane 0 & 1 pll reset\n");
				ifrtp_set_tx_discard_all(ifrtp_cfg.driver_handle->dev[0], true);
				ifrtp_set_tx_discard_all(ifrtp_cfg.driver_handle->dev[1], true);

				if (ifrtp_fpga_pll_reset())
					pr_warn("PLL reset timedout\n");

				ifrtp_set_tx_discard_all(ifrtp_cfg.driver_handle->dev[0], false);
				ifrtp_set_tx_discard_all(ifrtp_cfg.driver_handle->dev[1], false);

				perform_pll_reset = false;

				// restart in IFRTP_LINK_UP
				memset(link_stat, 0, sizeof(link_stat));
				memset(mdio_reset_done, 0, sizeof(mdio_reset_done));
				memset(rx_datapath_reset_count, 0, sizeof(rx_datapath_reset_count));

				goto exit_link_sequence;
			}

			if (!mdio_reset_done[i]) {
				if (ifrtp_mdio_link_reset(i, ifrtp_cfg.mdio_reset_retry_interval_ms, ifrtp_cfg.mdio_reset_retry_times))
					pr_err("error reset pcs/pma through mdio %d\n", i);
				mdio_reset_done[i] = true;
				break;
			}

			if (!rx_datapath_reset_count[i]) {
				if (!ifrtp_toggle_rx_datapath_reset(1 << (IFRTP_PHY_COUNT - 1 - i)))
					rx_datapath_reset_count[i] = 1;

				break;
			}

			if (enter_normal_process[i]) {
				pr_info("WAN%d: enter link normal process\n", i);
				link_stat[i] = IFRTP_LINK_NORMAL;
				break;
			}

			if ((lane_status & IFRTP_SFP_STAT_LINK_OK_MASK) == IFRTP_SFP_STAT_LINK_OK) {
				pr_info("WAN%d: lane_stat[18:16] = 3\n", i);
				enter_normal_process[i] = true;
				break;
			}

			if ((lane_status & IFRTP_SFP_STAT_LINK_OK_MASK) == (7 << 16) &&
			    ktime_get_boottime_seconds() < ESTIMATED_BOOTTIME_SECONDS) {
				pr_info("WAN%d: lane_stat[18:16] = 7\n", i);
				perform_pll_reset = true;
				break;
			}

			if (rx_datapath_reset_count[i] > 4) {
				link_stat[i] = IFRTP_LINK_DOWN;
				pr_warn("WAN%d: failed to bring link up after 5 rx_datapath_reset, enter link down process\n", i);
			} else {
				if (!ifrtp_toggle_rx_datapath_reset(1 << (IFRTP_PHY_COUNT - 1 - i)))
					++rx_datapath_reset_count[i];
			}
			break;
		case IFRTP_LINK_NORMAL:
			if (!ifrtp_sfp_is_linked_up(i) || (ifrtp_i2c_module_read(i) & IFRTP_I2C_BIT_SFP_RX_LOS)) {
				link_stat[i] = IFRTP_LINK_DOWN;
				// TODO print more log for debuging
				pr_info("WAN%d: stat_rx_status != 3\n", i);
				pr_info("WAN%d: enter link down process\n", i);
			}
			break;
		case IFRTP_LINK_DOWN:
			if (lane_status & IFRTP_SFP_STAT_RX_STATUS) {
				pr_info("WAN%d: stat_rx_status = 1\n", i);
				link_stat[i] = IFRTP_LINK_UP;
				enter_normal_process[i] = false;
				rx_datapath_reset_count[i] = 0;
				pr_info("WAN%d: enter link up process\n", i);
			} else if (!(ifrtp_i2c_module_read(i) & IFRTP_I2C_BIT_SFP_RX_LOS)) {
				ifrtp_pr_debug("WAN%d: RX_LOS = 0\n", i);
				ifrtp_toggle_rx_datapath_reset(1 << (IFRTP_PHY_COUNT - 1 - i));
			} else {
				ifrtp_pr_debug("WAN%d: RX_LOS = 1\n", i);
				mdio_reset_done[i] = false;
			}
			break;
		}
	}

exit_link_sequence:
	ifrtp_update_packet_count();
	ifrtp_update_firewall_settings();
}

void ifrtp_init(struct net_device *dev)
{
	struct ifrtp_priv *priv;

	dev->type		= ARPHRD_NONE;
	dev->hard_header_len	= 0;
	dev->min_header_len	= 0;
	dev->addr_len		= 0;
	dev->flags		= IFF_POINTOPOINT | IFF_NOARP;
	dev->mtu		= ETH_MAX_MTU;
	dev->min_mtu		= ETH_MIN_MTU;
	dev->max_mtu		= ETH_MAX_MTU;
	dev->tx_queue_len	= 1000; // DEFAULT_TX_QUEUE_LEN
	dev->watchdog_timeo	= 5;
	dev->netdev_ops		= &ifrtp_netdev_ops;
	dev->ethtool_ops	= &ifrtp_ethtool_ops;
	dev->features		|= NETIF_F_HW_CSUM;

	priv = netdev_priv(dev);
	memset(priv, 0, sizeof(struct ifrtp_priv));
	priv->dev = dev;
}


int ifrtp_notifier_call_inetaddr(struct notifier_block *nb, unsigned long event, void *data)
{
	struct in_ifaddr *ifaddr = data;
	struct IocNetTxSourceIpAddressConfig ip_config = {
		.ch_no = 0,
		.ip_dst_addr = 0,
	};

	if (event == NETDEV_UP || event == NETDEV_DOWN) {
		if (event == NETDEV_UP)
			ip_config.ip_dst_addr = ifaddr->ifa_address;

		if (memcmp(ifaddr->ifa_dev->dev->name, IGE_IFNAME"0", sizeof(IGE_IFNAME)) == 0)
			ifrtp_set_tx_source_primary_address(NULL, &ip_config);
		else if (memcmp(ifaddr->ifa_dev->dev->name, IGE_IFNAME"1", sizeof(IGE_IFNAME)) == 0)
			ifrtp_set_tx_source_secondary_address(NULL, &ip_config);
		else
			return 0;
	}

	if (event == NETDEV_UP) {
		// update target mac addr on ip addr change
		queue_delayed_work(arp_check_wq, &ifrtp_work_arp_check, HZ * 4);
	}
	return 0;
}

static struct notifier_block nb = {
	.notifier_call = ifrtp_notifier_call_inetaddr,
};

static bool ifrtp_dma_channel_match_name(struct dma_chan *chan, void *param)
{
	if (!param || ((char *)param)[0] == '\0')
		return true;
	return strcmp(dma_chan_name(chan), param) == 0;
}

void ifrtp_init_rtp_recv_buffer(struct ifrtp_priv *priv)
{
	int i, j;
	int media_type = priv->stream_id * RTP_SSRC_PER_CH;

	ifrtp_dev_dbg(&priv->dev->dev, "media_type = %d\n", media_type);
	for (i = 0; i < RTP_SSRC_PER_CH; ++i) {
		int media = i + priv->stream_mode * IFRTP_STREAM_BUFFER_MEMAP_NUM_STREAM_TYPE;

		for (j = 0; j < priv->ssrc_count[i]; ++j) {
			priv->recv_buff[media_type].begin = priv->recv_buff[media_type].begin_offset +
				ifrtp_cfg.rtp_recv_buffer[media].start_addr +
				ifrtp_cfg.rtp_recv_buffer[media].ch_offset * (priv->stream_id + j);
			priv->recv_buff[media_type].end = priv->recv_buff[media_type].begin +
				ifrtp_cfg.rtp_recv_buffer[media].size - RTP_PACKET_OFFSET_SIZE;
			priv->recv_buff[media_type].addr = priv->recv_buff[media_type].begin;

			ifrtp_dev_dbg(&priv->dev->dev, "i = %d, j = %d\n", i, j);
			ifrtp_dev_dbg(&priv->dev->dev, "recv_buff[%d].begin = 0x%08lx\n", media_type, priv->recv_buff[media_type].begin);
			ifrtp_dev_dbg(&priv->dev->dev, "recv_buff[%d].end   = 0x%08lx\n", media_type, priv->recv_buff[media_type].end);
			++media_type;
		}
	}
}

#ifdef IFRTP_ENABLE_DEBUGFS
void ifrtp_init_debugfs(void)
{
	ifrtp_debugfs_dir = debugfs_create_dir("ifrtp", NULL);

	debugfs_arrary_1.n_elements = 2048;
	debugfs_arrary_1.array = kmalloc_array(debugfs_arrary_1.n_elements, sizeof(u32), GFP_KERNEL);
	debugfs_create_u32_array("ifrtp_debugfs_arrary_1", 0444, ifrtp_debugfs_dir, &debugfs_arrary_1);

	debugfs_create_u32("ifrtp_debugfs_counter_1",
			0664,
			ifrtp_debugfs_dir,
			&debugfs_counter_1);
	debugfs_create_u32("trigger_mdio_core_reset",
			0664,
			ifrtp_debugfs_dir,
			&debugfs_trigger_mdio_core_reset);
	debugfs_create_x32("ifrtp_debugfs_i2c_0",
			0444,
			ifrtp_debugfs_dir,
			&debugfs_i2c_module_0);
	debugfs_create_x32("ifrtp_debugfs_i2c_1",
			0444,
			ifrtp_debugfs_dir,
			&debugfs_i2c_module_1);
	debugfs_create_x32("ifrtp_debugfs_mdio_0",
			0664,
			ifrtp_debugfs_dir,
			&debugfs_mdio_0);
	debugfs_create_x32("ifrtp_debugfs_mdio_1",
			0664,
			ifrtp_debugfs_dir,
			&debugfs_mdio_1);
	debugfs_create_bool("ifrtp_debugfs_link_no_autoreset",
			0664,
			ifrtp_debugfs_dir,
			&debugfs_link_no_autoreset);
	debugfs_create_bool("ifrtp_debugfs_link_sw_force_up",
			0664,
			ifrtp_debugfs_dir,
			&debugfs_link_sw_force_up);
	debugfs_create_u32("ifrtp_timeout_ioctl_counter",
			0664,
			ifrtp_debugfs_dir,
			&ifrtp_cfg.timeout_ioctl_debugfs_counter);
	debugfs_create_x32("ifrtp_debugfs_var_1",
			0664,
			ifrtp_debugfs_dir,
			&debugfs_var_1);
}
#endif

extern int ifrtp_ioc_set_if_dest_addr(struct ifrtp_priv *priv, char *ioc_data);
static struct workqueue_struct *netlink_reply_wq;
struct sock *ifrtp_nl_sk;

struct ifrtp_netlink_req {
	struct work_struct work;
	struct ifreq req;
	int pid;
};

static void ifrtp_reply_netlink(struct work_struct *work)
{
	struct ifrtp_netlink_req *work_data = container_of(work, struct ifrtp_netlink_req, work);
	struct sockaddr *sa = &work_data->req.ifr_dstaddr;
	struct sockaddr_in *sin = (struct sockaddr_in *)sa;
	struct sk_buff *skb_out;
	struct nlmsghdr *nlh;
	int res, if_index;

	ifrtp_pr_debug("Processing ifreq ifname:%s family: %d\n", work_data->req.ifr_name, sa->sa_family);
	if (sa->sa_family == AF_INET)
		ifrtp_pr_debug("addr: %pI4:%d\n", &sin->sin_addr, ntohs(sin->sin_port));

	skb_out = nlmsg_new(sizeof(int), 0);
	if (!skb_out) {
		pr_err("Failed to allocate new skb for sending netlink response\n");
		kfree(work);
		return;
	}

	res = sscanf(work_data->req.ifr_name, "ivfrtp%d", &if_index);
	if (res != 1) {
		pr_err("Error matching netlink request interface: %s\n", work_data->req.ifr_name);
		res = -EINVAL;
		goto netlink_reply;
	}

	if (if_index < 0 || if_index >= ifrtp_cfg.numvfs) {
		pr_err("Invalid ivfrtp interface index: ivfrtp%d\n", if_index);
		res = -EINVAL;
		goto netlink_reply;
	}

	res = ifrtp_ioc_set_if_dest_addr(netdev_priv(ifrtp_devs[if_index]), (char *)sa);

netlink_reply:
	nlh = nlmsg_put(skb_out, 0, 0, NLMSG_DONE, sizeof(int), 0);
	if (!nlh) {
		pr_err("Failed to put reply message into skb\n");
		nlmsg_free(skb_out);
		kfree(work);
		return;
	}
	*((int *)nlmsg_data(nlh)) = res;

	res = nlmsg_unicast(ifrtp_nl_sk, skb_out, work_data->pid);

	if (res < 0)
		pr_err("Error sending netlink reply back to user: %d\n", res);
	else
		ifrtp_pr_debug("netlink reply finished.\n");

	kfree(work);
}

static void ifrtp_nl_recv_msg(struct sk_buff *skb)
{
	struct ifrtp_netlink_req *work;
	struct nlmsghdr *nlh = (struct nlmsghdr *)skb->data;
	struct ifreq *req = (struct ifreq *)nlmsg_data(nlh);
	struct sockaddr_in *sin = (struct sockaddr_in *)&req->ifr_dstaddr;

	ifrtp_pr_debug("Netlink received ifreq %d bytes, ifname:%s, family: %d, addr: %pI4:%d\n",
			nlh->nlmsg_len,
			req->ifr_name,
			req->ifr_dstaddr.sa_family,
			&sin->sin_addr, ntohs(sin->sin_port));

	work = kmalloc(sizeof(struct ifrtp_netlink_req), GFP_KERNEL);
	if (!work)
		return;

	INIT_WORK((struct work_struct *)work, ifrtp_reply_netlink);
	work->pid = nlh->nlmsg_pid;
	memcpy(&work->req, req, sizeof(struct ifreq));

	queue_work(netlink_reply_wq, (struct work_struct *)work);
}

static struct netlink_kernel_cfg ifrtp_netlink_cfg = {
	.input = ifrtp_nl_recv_msg,
};

int ifrtp_init_netlink(struct net_device *dev)
{
	struct net *ns_dev = dev_net(dev);

	ifrtp_nl_sk = netlink_kernel_create(ns_dev, NETLINK_IFRTP, &ifrtp_netlink_cfg);
	if (unlikely(!ifrtp_nl_sk)) {
		netdev_err(dev, "Error creating netlink socket.\n");
		return -ENOMEM;
	}

	netlink_reply_wq = create_singlethread_workqueue("ifrtp_nlrp");
	if (unlikely(!netlink_reply_wq)) {
		netdev_err(dev, "Error creating netlink reply workqueue.\n");
		return -ENOMEM;
	}

	return 0;
}

int ifrtp_init_driver(struct platform_device *pdev)
{
	int i, j, ret = -ENOMEM;

#ifdef IFRTP_ENABLE_DEBUGFS
	ifrtp_init_debugfs();
#endif

	ifrtp_cfg.pdev = pdev;
	for (i = 0; i < ifrtp_cfg.numvfs; ++i) {
		ifrtp_devs[i] = alloc_netdev(sizeof(struct ifrtp_priv), "ivfrtp%d",
				NET_NAME_UNKNOWN, ifrtp_init);
		if (ifrtp_devs[i] == NULL)
			goto error_out;
	}

#if KERNEL_VERSION(6, 2, 0) <= LINUX_VERSION_CODE
	netdev_sw_irq_coalesce_default_on(ifrtp_devs[0]);
#else
	ifrtp_devs[0]->gro_flush_timeout = 200000;
	ifrtp_devs[0]->napi_defer_hard_irqs = 2;
#endif

	// separate napi processing from ksoftirqd
	dev_set_threaded(ifrtp_devs[0], true);

	for (i = 0; i < ifrtp_cfg.numvfs; i++) {
		ret = register_netdev(ifrtp_devs[i]);
		if (ret) {
			dev_err(&pdev->dev, "error %i registering device %s\n", ret, ifrtp_devs[i]->name);
			goto error_out;
		}
	}

#if KERNEL_VERSION(6, 1, 0) <= LINUX_VERSION_CODE
	netif_napi_add_weight(ifrtp_devs[0], &ifrtp_napi, ifrtp_poll, ifrtp_cfg.rtp_poll_weight);
#else
	netif_napi_add(ifrtp_devs[0], &ifrtp_napi, ifrtp_poll, ifrtp_cfg.rtp_poll_weight);
#endif

	ret = dma_set_mask_and_coherent(&pdev->dev, DMA_BIT_MASK(64));
	if (unlikely(ret)) {
		dev_err(&pdev->dev, "error set dma mask: %d\n", ret);
		goto error_out;
	}

	dma_cap_mask_t mask;

	dma_cap_zero(mask);
	dma_cap_set(DMA_MEMCPY, mask);

#ifndef IFRTP_RX_USE_DMAC_MEMCPY
	rtp_recv_buffer_remaped = ioremap_wc(ifrtp_cfg.rtp_recv_buffer[0].start_addr, 0x09100000);
	rtp_recv_buffer_remap_offset = rtp_recv_buffer_remaped - ifrtp_cfg.rtp_recv_buffer[0].start_addr;
#else
	dmachan_memcpy = dma_request_channel(mask, ifrtp_dma_channel_match_name, IFRTP_DMA_CHANNEL_1);
	if (!dmachan_memcpy) {
		dev_warn(&pdev->dev, "can't allocate channel %s\n", IFRTP_DMA_CHANNEL_1);
		return -EBUSY;
	}
#endif

	// prepare tx sk_buff memory
	tx_skbuff_p = kmalloc(ifrtp_cfg.tx_buffer * sizeof(struct sk_buff *), GFP_KERNEL);
	if (!tx_skbuff_p)
		goto error_out;

	// prepare tx descriptor memory
	desc_addr_tx = kmalloc(IFRTP_TX_DESCRIPTOR_DEPTH * 16, GFP_KERNEL | GFP_DMA);
	ifrtp_pr_debug("desc_addr_tx = 0x%px\n", desc_addr_tx);
	if (!desc_addr_tx)
		goto error_out;

	desc_addr_tx_phys = dma_map_single(&pdev->dev, desc_addr_tx,
			IFRTP_TX_DESCRIPTOR_DEPTH * 16, DMA_TO_DEVICE);
	if (unlikely(desc_addr_tx_phys == DMA_MAPPING_ERROR)) {
		dev_err(&pdev->dev, "error mapping dma memory for tx descriptor\n");
		goto error_out;
	}
	ifrtp_pr_debug("dma: desc_addr_tx      = 0x%px\n", desc_addr_tx);
	ifrtp_pr_debug("dma: desc_addr_tx_phys = 0x%016llx\n", desc_addr_tx_phys);
	ifrtp_iow32(NULL, IFRTP_TX_DESCRIPTOR_DEPTH, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_DEPTH);
	ifrtp_iow32(NULL, desc_addr_tx_phys >> 32, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_DAH);
	ifrtp_iow32(NULL, desc_addr_tx_phys & 0xFFFFFFFF, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_DAL);
	ifrtp_iow32(NULL, 0, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_WP);

	// prepare rx packet memory
	packet_addr_rx = kmalloc(IFRTP_RX_DESCRIPTOR_DEPTH * 2048, GFP_KERNEL | GFP_DMA);
	ifrtp_pr_debug("packet_addr_rx = 0x%px\n", packet_addr_rx);
	if (!packet_addr_rx)
		goto error_out;

	packet_addr_rx_phys = dma_map_single(&pdev->dev, packet_addr_rx,
			IFRTP_RX_DESCRIPTOR_DEPTH * 16, DMA_TO_DEVICE);
	if (unlikely(packet_addr_rx_phys == DMA_MAPPING_ERROR)) {
		dev_err(&pdev->dev, "error mapping dma memory for rx packet buffer\n");
		goto error_out;
	}

	// prepare rx descriptor memory
	desc_addr_rx = kmalloc(IFRTP_RX_DESCRIPTOR_DEPTH * 16, GFP_KERNEL | GFP_DMA);
	ifrtp_pr_debug("desc_addr_rx = 0x%px\n", desc_addr_rx);
	if (!desc_addr_rx)
		goto error_out;

	for (i = 0; i < IFRTP_RX_DESCRIPTOR_DEPTH; ++i) {
		desc_addr_rx[i*4]	= (packet_addr_rx_phys + i*2048) >> 32;
		desc_addr_rx[i*4+1]	= (packet_addr_rx_phys + i*2048) & 0xFFFFFFFF;
	}

	desc_addr_rx_phys = dma_map_single(&pdev->dev, desc_addr_rx,
			IFRTP_RX_DESCRIPTOR_DEPTH * 16, DMA_TO_DEVICE);
	if (unlikely(desc_addr_rx_phys == DMA_MAPPING_ERROR)) {
		dev_err(&pdev->dev, "error mapping dma memory for rx descriptor\n");
		goto error_out;
	}
	ifrtp_iow32(NULL, IFRTP_RX_DESCRIPTOR_DEPTH, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_DEPTH);
	ifrtp_iow32(NULL, desc_addr_rx_phys >> 32, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_DAH);
	ifrtp_iow32(NULL, desc_addr_rx_phys & 0xFFFFFFFF, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_DAL);
	ifrtp_iow32(NULL, 0, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_RP);

	for (i = 0; i < ifrtp_cfg.numvfs; i++) {
		struct ifrtp_priv *priv = netdev_priv(ifrtp_devs[i]);

		priv->tx_desc_lock		= &tx_desc_lock;
		priv->desc_addr_tx		= desc_addr_tx;
		priv->desc_addr_rx		= desc_addr_rx;
		priv->stream_id			= i / 2;
		priv->is_main_mac		= !(i & 1);
		priv->stream_mode		= 0;
		priv->ssrc_mismatch_last	= 0;
		priv->recv_buff			= g_recv_buff;
		for (j = 0; j < RTP_SSRC_PER_CH; ++j)
			priv->ssrc_count[j] = 1;

		ifrtp_init_rtp_recv_buffer(priv);
	}

	ifrtp_init_packet_counter();

	// setup workqueue for sfp rx_loss and packet counter
	if (!ifrtp_2hz_periodic_wq)
		ifrtp_2hz_periodic_wq = create_singlethread_workqueue("ifrtp_2hz");
	if (ifrtp_2hz_periodic_wq)
		queue_delayed_work(ifrtp_2hz_periodic_wq, &ifrtp_work_2hz_periodic, HZ * ESTIMATED_BOOTTIME_SECONDS / 2);
	else
		dev_err(&pdev->dev, "Error creating workqueue.\n");

	// setup tx arp check workqueue
	if (!arp_check_wq)
		arp_check_wq = create_singlethread_workqueue("ifrtp_arp");
	if (arp_check_wq)
		queue_delayed_work(arp_check_wq, &ifrtp_work_arp_check, HZ * ESTIMATED_BOOTTIME_SECONDS);
	else
		dev_err(&pdev->dev, "Error creating arp check workqueue.\n");

	if (register_inetaddr_notifier(&nb))
		dev_warn(&pdev->dev, "Can't register inetaddr notifier.\n");

	// Not actually used, just a placeholder for someone else. (maybe RTOS?)
	struct dma_chan *dmachan = dma_request_channel(mask, ifrtp_dma_channel_match_name, IFRTP_DMA_CHANNEL_0);

	if (!dmachan)
		dev_warn(&pdev->dev, "Can't allocate DMA channel %s\n", IFRTP_DMA_CHANNEL_0);

	napi_enable(&ifrtp_napi);
	ifrtp_enable_net_rx_irq();

error_out:
	if (ret)
		ifrtp_cleanup_driver();
	return ret;
}

void ifrtp_cleanup_driver(void)
{
	int i;

	pr_info("cleaning up ifrtp...\n");

	if (ifrtp_nl_sk)
		netlink_kernel_release(ifrtp_nl_sk);

	if (netlink_reply_wq) {
		flush_workqueue(netlink_reply_wq);
		destroy_workqueue(netlink_reply_wq);
	}

	if (arp_check_wq) {
		cancel_delayed_work_sync(&ifrtp_work_arp_check);
		flush_workqueue(arp_check_wq);
		destroy_workqueue(arp_check_wq);
	}

	if (ifrtp_2hz_periodic_wq) {
		cancel_delayed_work_sync(&ifrtp_work_2hz_periodic);
		flush_workqueue(ifrtp_2hz_periodic_wq);
		destroy_workqueue(ifrtp_2hz_periodic_wq);
	}

	napi_disable(&ifrtp_napi);
	for (i = 0; i < IFRTP_NUMVFS_MAX; i++) {
		if (ifrtp_devs[i]) {
			unregister_netdev(ifrtp_devs[i]);
			free_netdev(ifrtp_devs[i]);
		}
	}

	unregister_inetaddr_notifier(&nb);

	if (tx_skbuff_p) {
		for (i = 0; i < ifrtp_cfg.tx_buffer; ++i) {
			consume_skb(tx_skbuff_p[i]);
		}
		kfree(tx_skbuff_p);
		tx_skbuff_p = NULL;
	}

#ifndef IFRTP_RX_USE_DMAC_MEMCPY
	iounmap(rtp_recv_buffer_remaped);
#else
	if (dmachan_memcpy)
		dma_release_channel(dmachan_memcpy);
#endif

#ifdef IFRTP_ENABLE_DEBUGFS
	debugfs_remove_recursive(ifrtp_debugfs_dir);
#endif
}

// query mac addr for paddr, using interface igw[ifrtp_if_id], and copy to eth_dst
int arp_query(__be32 paddr, int igw_if_id, char *eth_dst)
{
	struct rtable *rt;
	struct neighbour *n;
	struct net_device *dev;
	struct net *ndev;
	struct in_ifaddr *ifa;
	__be32 haddr, saddr;
	int ret = -EHOSTUNREACH;

	if (paddr == 0)
		return 0;

	ifrtp_pr_debug("ARP Querying IP [%pI4] from igw%d\n", &paddr, igw_if_id);

	dev = ifrtp_cfg.driver_handle->netdev[igw_if_id];
	if (!dev || !(dev->flags & IFF_UP) || !netif_carrier_ok(dev)) {
		pr_warn("igw%d not available for arp\n", igw_if_id);
		return ret;
	}

	rcu_read_lock();
	ifa = rcu_dereference(dev->ip_ptr->ifa_list);
	if (ifa) {
		saddr = ifa->ifa_address;
	} else {
		netdev_warn(dev, "no available source ip addr\n");
		saddr = 0;
	}
	rcu_read_unlock();

	ndev = dev_net(dev);
	rt = ip_route_output(ndev, paddr, saddr, RT_TOS(0), dev->ifindex);
	if (IS_ERR(rt)) {
		netdev_err(dev, "error reading routing table\n");
		return PTR_ERR(rt);
	}

	if (rt->rt_uses_gateway) {
		if (unlikely(rt->rt_gw_family != AF_INET)) {
			netdev_err(dev, "arp requires IPv4 to work, rt_gw_family: 0x%02x\n", rt->rt_gw_family);
			return ret;
		}
		haddr = rt->rt_gw4; // default gateway ip
		ifrtp_pr_debug("arp using Default Gateway IP [%pI4]\n", &haddr);
	} else {
		ifrtp_pr_debug("arp rt_uses_gateway: %d\n", rt->rt_uses_gateway);
		haddr = paddr;
	}

	ifrtp_pr_debug("%s haddr = %pI4\n", __func__, &haddr);
	n = __ipv4_neigh_lookup(dev, haddr);
	if (!n)
		n = neigh_create(&arp_tbl, &haddr, dev);

	if (!IS_ERR(n)) {
		ret = 0;
		ifrtp_pr_debug("Target MAC [%pM], nud 0x%02x\n", n->ha, n->nud_state);
		if (neigh_event_send(n, NULL)) {
			int timeout = ifrtp_cfg.arp_timeout_ms;
			int sleep_ms = 10;

			ifrtp_pr_debug("neigh event sent\n");
			while ((n->nud_state & NUD_INCOMPLETE) && timeout > 0) {
				msleep(sleep_ms);
				ifrtp_pr_debug("sleep %d ms\n", sleep_ms);
				ifrtp_pr_debug("Target MAC [%pM], nud 0x%02x\n", n->ha, n->nud_state);
				timeout -= sleep_ms;
				sleep_ms *= 2;
			}
			if ((timeout <= 0) || !(n->nud_state & NUD_VALID)) {
				netdev_warn(dev, "timeout in arp querying %pI4 from %s mac\n", &paddr, igw_if_id ? "sub" : "main");
				ret = -EHOSTUNREACH;
			} else {
				ret = 0;
			}
		}
		ifrtp_pr_debug("Target MAC [%pM], nud 0x%02x\n", n->ha, n->nud_state);
		memcpy(eth_dst, n->ha, sizeof(char[6]));
		neigh_release(n);
		return ret;
	}

	netdev_err(dev, "error lookup ipv4 neighbour\n");
	return PTR_ERR(n);
}

