// SPDX-License-Identifier: GPL-2.0
/*
 * Sony ifrtp SFP Ethernet device driver
 *
 * Copyright (c) 2023 Sony Corporation. All rights reserved.
 */
#include <linux/kern_levels.h>
#include <linux/delay.h>
#include "ifrtp.h"
#include "ifrtp_mdio.h"
#include "io.h"
#include "igw_net_address.h"

#define MDIO_OP_CL22_WRITE	0x50000000
#define MDIO_OP_CL22_READ	0x60000000
#define MDIO_OP_RDY		       0x2
#define MDIO_OP_GO		       0x1
#define MDIO_REG_ADDR_SB	2
#define MDIO_PHY_ADDR_SB	7
#define MDIO_PHY_ADDR_BASE	4
#define MDIO_CL22_DATA_SB	12
#define MDIO_CL22_DATA_MASK	0x0FFFF000

#define MDIO_OP_RDY_TIMEOUT_US	200000
#define MDIO_OP_RDY_SLEEP_US	10

static DEFINE_MUTEX(mdio_interface_lock);
int __mdio_write16(uint8_t addr, uint16_t data, int area_no)
{
	if (addr > 0x1F) {
		pr_err("invalid mdio address 0x%02x", addr);
		return -EINVAL;
	}
	if (area_no > 1) {
		pr_err("invalid mdio area_no %d", area_no);
		return -EINVAL;
	}

	unsigned int access_data = MDIO_OP_CL22_WRITE |
					((uint32_t)data << MDIO_CL22_DATA_SB) |
					((MDIO_PHY_ADDR_BASE) << MDIO_PHY_ADDR_SB) |
					((uint32_t)addr << MDIO_REG_ADDR_SB) |
					MDIO_OP_RDY | MDIO_OP_GO;
	unsigned int offset_area_top = OFFSET_NET_GMAC_TOP + OFFSET_NET_GMAC_CTL_TOP +
					(area_no * OFFSET_NET_GMAC_CTL_SIZE);
	unsigned int address = offset_area_top + OFFSET_NET_GMAC_CTL_MDIO_ACCESS;

	ifrtp_iow32(NULL, access_data, address);

	uint32_t res;
	int err = read_poll_timeout(ifrtp_ior32, res, (res & MDIO_OP_RDY),
			MDIO_OP_RDY_SLEEP_US, MDIO_OP_RDY_TIMEOUT_US, false, NULL, address);

	return err;
}

int __mdio_read16(uint8_t addr, uint16_t *data, int area_no)
{
	if (addr > 0x1F) {
		pr_err("invalid mdio address 0x%02x", addr);
		return -EINVAL;
	}
	if (area_no > 1) {
		pr_err("invalid mdio area_no %d", area_no);
		return -EINVAL;
	}

	unsigned int access_data = MDIO_OP_CL22_READ |
					((MDIO_PHY_ADDR_BASE) << MDIO_PHY_ADDR_SB) |
					((uint32_t)addr << MDIO_REG_ADDR_SB) |
					MDIO_OP_RDY | MDIO_OP_GO;
	unsigned int offset_area_top = OFFSET_NET_GMAC_TOP + OFFSET_NET_GMAC_CTL_TOP +
					(area_no * OFFSET_NET_GMAC_CTL_SIZE);
	unsigned int address = offset_area_top + OFFSET_NET_GMAC_CTL_MDIO_ACCESS;

	ifrtp_iow32(NULL, access_data, address);

	uint32_t res;
	int err = read_poll_timeout(ifrtp_ior32, res, (res & MDIO_OP_RDY),
			MDIO_OP_RDY_SLEEP_US, MDIO_OP_RDY_TIMEOUT_US, false, NULL, address);
	*data = (res & MDIO_CL22_DATA_MASK) >> MDIO_CL22_DATA_SB;

	return err;
}


// refer to https://docs.xilinx.com/v/u/en-US/gig_eth_pcs_pma_ug155 12.2 Management Registers

bool ifrtp_mdio_sfp_is_linked_up(int interface)
{
	uint16_t data;

	// LinkStatus need to be read twice
	__mdio_read16(1, &data, interface);

	if (__mdio_read16(0, &data, interface))
		return false;

	if ((data & 0x0140) != 0x0140)
		return false;

	if (__mdio_read16(1, &data, interface))
		return false;

	if (data & 4)
		return true;
	else
		return false;
}

int ifrtp_mdio_link_reset(int interface, int timeout_ms, int retry)
{
	bool autoneg = ifrtp_cfg.driver_handle->dev[interface]->__netarch.eth_config.autoneg;
	u16 mdio_ctrl_reg = autoneg ? 0x1340 : 0x0140;

	pr_info("WAN%d: mdio reset link\n", interface);

	// Core reset
	int err = __mdio_write16(0, 0x8000, interface);

	if (err)
		return err;

	if (ifrtp_cfg.mdio_reset_pulse_ms)
		msleep(ifrtp_cfg.mdio_reset_pulse_ms);

	// Auto-Negotiation Enable
	// Auto-Negotiation Restart
	// Full-Duplex Mode
	// Speed Selection 10 (1000Mbps)
	pr_info("set mdio control register to 0x%04x\n", mdio_ctrl_reg);
	err = __mdio_write16(0, mdio_ctrl_reg, interface);
	return err;
}

void ifrtp_mdio_link_power_down(int interface)
{
	__mdio_write16(0, 0x0800, interface);
}
