// SPDX-License-Identifier: GPL-2.0
/*
 * Sony ifrtp SFP Ethernet device driver
 *
 * Copyright (c) 2023 Sony Corporation. All rights reserved.
 */
#include <linux/version.h>
#include <linux/netdevice.h>
#include <linux/if_ether.h>
#include <linux/etherdevice.h>
#include <linux/ethtool.h>
#include <linux/byteorder/generic.h>

#include "ifrtp_ioctl.h"
#include "ifrtp_mdio.h"
#include "ifrtp_driver.h"
#include "igw_driver.h"
#include "igw_netreg.h"
#include "igw_net_address.h"

#include "ifrtp.h"
#include "ifrtp_filter.h"
#include "io.h"


#define IGW_NET_LINKDETECTION_TIMER_TIMEOUT	(HZ >> 1) // 500msec
#define IGW_NET_MAX_PACKET_SIZE			(1024*64)

static void igw_timer_handler(struct timer_list *t);

static bool check_packet_receive(struct igw_priv *dev, bool is_init);

#define LINK_STATE_DOWN			0x00
#define LINK_STATE_UNKNOWN		0x01
#define LINK_STATE_UP			0x02
#define LINK_STATE_UP_PACKET_RECEIVED	0x04


static bool is_open_tx_discard[IFRTP_PHY_COUNT];
static bool is_open_rx_discard[IFRTP_PHY_COUNT];
static bool is_open_mii_filter[IFRTP_PHY_COUNT];

static void open_tx_discard(struct net_device *dev)
{
	struct igw_priv *priv = (struct igw_priv *) netdev_priv(dev);

	if (is_open_tx_discard[priv->netif_index])
		return;

	spin_lock(&priv->__netarch.lock);

	ifrtp_set_tx_discard_all(priv, false);
	ifrtp_set_tx_discard_by_devindex(priv, false);

	is_open_tx_discard[priv->netif_index] = true;

	spin_unlock(&priv->__netarch.lock);
}

static void open_rx_discard(struct net_device *dev)
{
	struct igw_priv *priv = (struct igw_priv *) netdev_priv(dev);

	if (is_open_rx_discard[priv->netif_index])
		return;

	spin_lock(&priv->__netarch.lock);

	ifrtp_set_rx_discard_all(priv, false);
	ifrtp_set_rx_discard_by_devindex(priv, false);

	is_open_rx_discard[priv->netif_index] = true;

	spin_unlock(&priv->__netarch.lock);
}

static void open_mii_filter(struct net_device *dev)
{
	struct igw_priv *priv = (struct igw_priv *) netdev_priv(dev);

	if (is_open_mii_filter[priv->netif_index])
		return;

	spin_lock(&priv->__netarch.lock);

	// packet受信検知できるように初期化(現状値を取得)しておく.
	check_packet_receive(priv, true);

	// TODO ? enable ethernet mii path

	is_open_mii_filter[priv->netif_index] = true;

	spin_unlock(&priv->__netarch.lock);
}

static void close_mii_filter(struct net_device *dev)
{
	struct igw_priv *priv = (struct igw_priv *) netdev_priv(dev);

	if (!is_open_mii_filter[priv->netif_index])
		return;

	// TODO ? disable ethernet mii path

	is_open_mii_filter[priv->netif_index] = false;
}

static int igw_open(struct net_device *dev)
{
	struct igw_priv *priv = (struct igw_priv *) netdev_priv(dev);
	static bool initilize[IFRTP_PHY_COUNT];

	// 最初の openの場合のみadd_timerして、ポーリング関数を呼び出すようにする.
	if (!initilize[priv->netif_index]) {
		add_timer(&priv->__netarch.link_check_timer);
		initilize[priv->netif_index] = true;
	}

	return 0;
}

static int igw_close(struct net_device *dev)
{
	netif_stop_queue(dev);
	pr_debug("%s netif stop queuing\n", dev->name);

	// mdio_link_power_down?
	// close_discard?
	// cannot bring up the interface again if close_discard
	// TODO (guo): turn off phy properly and fix ifrtp_open

	return 0;
}

static void igw_get_stats64(struct net_device *dev, struct rtnl_link_stats64 *storage)
{
	struct igw_priv *priv = (struct igw_priv *) netdev_priv(dev);

	*storage = priv->__netarch.stats;
}

static int igw_send(struct sk_buff *skb, struct net_device *dev)
{
	int result = 0;
	struct igw_priv *priv;
	unsigned int type = 0;

	priv = (struct igw_priv *) netdev_priv(dev);

	type = TX_TYPE_MAIN + (priv->netif_priority);

	if (unlikely(skb->len <= 0 || skb->data == NULL)) {
		result = -EINVAL;
		goto IGW_SEND_FIN;
	}

	// pad skb if skb->len < 60
	if (eth_skb_pad(skb)) {
		result = -ENOMEM;
		goto IGW_SEND_FIN;
	}

	if (ntohs(skb->protocol) == ETH_P_IP && ip_hdr(skb)->protocol == SOL_UDP) {
		// accept ephemeral port will always success
		// the oldest filter for ephemeral port will be reused
		ifrtp_filter_accept_ephemeral_port(priv, ntohs(udp_hdr(skb)->source));
	}

	// XXX: workaround for FPGA aligner
	if ((uintptr_t)skb->data & 0xF && skb->len & 0xF) {
		size_t pad_size = 0x10 - (skb->len & 0xF);
		if (skb_tailroom(skb) >= pad_size)
			memset(skb_put(skb, pad_size), 0, pad_size);
	}

	netif_trans_update(dev);

	struct ifrtp_msghdr msg = {
		.msg_addr = (uintptr_t)skb->data,
		.msg_len = skb->len,
		.ssrc = 0,
	};

	result = ifrtp_send_direct_multi_message((unsigned char *)&msg, sizeof(struct ifrtp_msghdr), type + (1 << 16), skb);

IGW_SEND_FIN:

	// 統計情報の更新
	if (likely(result >= 0)) {
		priv->__netarch.stats.tx_packets++;
		priv->__netarch.stats.tx_bytes += msg.msg_len;
	} else {
		pr_warn("%s failed to send packet(code:%d)\n", dev->name, result);
		switch (result) {
		case -EAGAIN:
			priv->__netarch.stats.tx_fifo_errors++;
			// tx_filo_errors should also be count as tx_errors, fallthrough intentionally
		case -EINVAL:
			priv->__netarch.stats.tx_errors++;
			break;
		case -ENOMEM:
			priv->__netarch.stats.tx_dropped++;
			break;
		}
	}

	return NETDEV_TX_OK;
}

static int igw_set_mac_address(struct net_device *dev, void *addr)
{
	struct sockaddr *mac = (struct sockaddr *) addr;

	memcpy(dev->dev_addr, mac->sa_data, IFHWADDRLEN);
	pr_debug("%s set mac %02x:%02x:%02x:%02x:%02x:%02x\n",
			dev->name,
			dev->dev_addr[0],
			dev->dev_addr[1],
			dev->dev_addr[2],
			dev->dev_addr[3],
			dev->dev_addr[4],
			dev->dev_addr[5]
	);

	struct IocNetTxSourceMacAddressConfig mac_config;

	memcpy(mac_config.eth_src, mac->sa_data, sizeof(mac_config.eth_src));

	if (memcmp(dev->name, IGE_IFNAME"0", sizeof(IGE_IFNAME)) == 0)
		ifrtp_set_tx_source_primary_mac_address(netdev_priv(dev), &mac_config);
	else
		ifrtp_set_tx_source_secondary_mac_address(netdev_priv(dev), &mac_config);

	return 0;
}

static int igw_get_link_ksettings(struct net_device *dev, struct ethtool_link_ksettings *elk)
{
	struct ethtool_link_settings *const ecmd = &elk->base;
	int result = 0;
	struct igw_priv *priv = (struct igw_priv *) netdev_priv(dev);
	bool is_link_up = netif_carrier_ok(dev);

	spin_lock(&priv->__netarch.lock);

	ecmd->transceiver = XCVR_EXTERNAL;
	ecmd->port = PORT_FIBRE;
	ecmd->autoneg = priv->__netarch.eth_config.autoneg;
	ecmd->duplex = (is_link_up ? DUPLEX_FULL : DUPLEX_UNKNOWN);
	// upしていれば、serdes rate, Downなら unknown.

	ethtool_link_ksettings_add_link_mode(elk, supported, FIBRE);
	// TODO add speed detect
	ecmd->speed = (__u32)(is_link_up ? SPEED_1000 : SPEED_UNKNOWN);

	spin_unlock(&priv->__netarch.lock);

	return result;
}

static int igw_set_link_ksettings(struct net_device *dev, const struct ethtool_link_ksettings *elk)
{
	const struct ethtool_link_settings *const ecmd = &elk->base;
	int result = 0;
	bool link_reset = false;
	struct igw_priv *priv = (struct igw_priv *) netdev_priv(dev);

	spin_lock(&priv->__netarch.lock);

	if (ecmd->autoneg != priv->__netarch.eth_config.autoneg) {
		switch (ecmd->autoneg) {
		case AUTONEG_ENABLE:
		case AUTONEG_DISABLE:
			priv->__netarch.eth_config.autoneg = ecmd->autoneg;
			link_reset = true;
			break;
		default:
			netdev_err(dev, "Could not set autodetect to value %d from ethtool\n", ecmd->autoneg);
			result = -EINVAL;
		}
	}

	if (!priv->__netarch.eth_config.autoneg) {
		if (ecmd->speed != priv->__netarch.eth_config.speed_target && ecmd->speed != SPEED_UNKNOWN) {
			switch (ecmd->speed) {
			case SPEED_1000:
				priv->__netarch.eth_config.speed_target = SPEED_1000;
				break;
			default:
				netdev_err(dev, "Failed to set ethernet speed to %dMbps, %dMbps is not supported\n", (int)ecmd->speed, (int)ecmd->speed);
				result = -EINVAL;
			}
		}
	}

	if (link_reset && result == 0)
		ifrtp_mdio_link_reset(priv->netif_priority, 0, 0);

	spin_unlock(&priv->__netarch.lock);
	return result;
}


static void igw_timer_handler(struct timer_list *t)
{
	struct igw_priv *priv = from_timer(priv, t, __netarch.link_check_timer);

	schedule_work((struct work_struct *)priv->__netarch.link_check_queue_work);

	mod_timer(&priv->__netarch.link_check_timer, jiffies + IGW_NET_LINKDETECTION_TIMER_TIMEOUT);
}


static bool check_packet_receive(struct igw_priv *dev, bool is_init)
{
	const unsigned int TIME_OUT_NO_PACKET_COUNT = 10;
	static unsigned int prev_recv_good_frame_cnt[IFRTP_PHY_COUNT];
	static unsigned int no_packet_count[IFRTP_PHY_COUNT];
	bool ret = is_init;
	unsigned int recv_good_frame_cnt = ifrtp_get_rx_xgmac_good_frame_lower(dev);

	if (is_init)
		no_packet_count[dev->netif_index] = 0;

	// 初期化でなくgoodframeが増えたかどうか.
	else if (recv_good_frame_cnt != prev_recv_good_frame_cnt[dev->netif_index])
		ret = true;

	// goodframe増えない.
	else {
		no_packet_count[dev->netif_index]++;
		if (no_packet_count[dev->netif_index] == TIME_OUT_NO_PACKET_COUNT) { // timeout.
			pr_err("IF[%d] Timed out with no packets received, interface is going to up.\n", dev->netif_index);
			ret = true;
		}
	}
	prev_recv_good_frame_cnt[dev->netif_index] = recv_good_frame_cnt;
	return ret;
}

static unsigned int get_phy_link_state(struct igw_priv *dev, unsigned int prev_phy_link_state)
{
	bool is_sfp_linked = ifrtp_sfp_is_linked_up(dev->netif_priority);
	bool is_serdes_linked = true;

	if (is_sfp_linked && is_serdes_linked) {
		return LINK_STATE_UP |
			(prev_phy_link_state & LINK_STATE_UP_PACKET_RECEIVED);
	}
	return LINK_STATE_DOWN;
}

static void igw_link_check_handler(struct work_struct *work)
{
	static unsigned int prev_phy_link_state[IFRTP_PHY_COUNT];

	struct ifrtp_link_check_queue_work *priv_work = (struct ifrtp_link_check_queue_work *)work;
	struct igw_priv *priv = priv_work->dev;
	unsigned int phy_link_state = get_phy_link_state(priv, prev_phy_link_state[priv->netif_index]);
	bool is_serdes_change_config = false;

	bool run = netif_running(priv->__netarch.dev);

	if (phy_link_state & LINK_STATE_UP && priv->__netarch.dev->flags & IFF_UP)
		ifrtp_led_on(priv->netif_priority);
	else
		ifrtp_led_off(priv->netif_priority);

	// UNKNOWN or 状態 or 設定 に変化があったどうか.
	if (phy_link_state & LINK_STATE_UNKNOWN ||
		phy_link_state != prev_phy_link_state[priv->netif_index] ||
		is_serdes_change_config
		) {
		// open中.
		if (run) {
			if (!netif_queue_stopped(priv->__netarch.dev)) {
				netif_stop_queue(priv->__netarch.dev);
				pr_info("%s netif stop queuing\n", priv->__netarch.dev->name);
			}
		}

		// 遮断させる.
		close_mii_filter(priv->__netarch.dev);
		// close_discard?

		if (netif_carrier_ok(priv->__netarch.dev))
			netif_carrier_off(priv->__netarch.dev);

		// serdesの構築(変更がなくてもEthrer Resetは実施させる).
		// link stateの更新.
		phy_link_state = get_phy_link_state(priv, phy_link_state);

		if (phy_link_state & LINK_STATE_UP) {
			// Upさせる.
			netif_carrier_on(priv->__netarch.dev);
			pr_info("IF[%d] Up\n", priv->netif_index);

			// mii filterまではrunning関係なく開けておく.
			open_mii_filter(priv->__netarch.dev);

			if (run) {
				// packetをkernelが送受信できる状態にする(RxはXGMACまで開通).
				open_tx_discard(priv->__netarch.dev);
				netif_wake_queue(priv->__netarch.dev);
				pr_info("%s netif wake queuing\n", priv->__netarch.dev->name);
			}
		} else if (phy_link_state & LINK_STATE_UNKNOWN) {
			if (!(prev_phy_link_state[priv->netif_index] & LINK_STATE_UNKNOWN)) { // 前回がUnKnown以外の場合.
				pr_info("IF[%d] Link Unknown\n", priv->netif_index);
			}
		} else if (phy_link_state == LINK_STATE_DOWN) {
			if (!(prev_phy_link_state[priv->netif_index] & LINK_STATE_DOWN)) { // 前回がDown以外の場合.
				pr_info("IF[%d] Down\n", priv->netif_index);
			}
		}
	}
	// Up.
	else if (phy_link_state & LINK_STATE_UP) {

		if (!(phy_link_state & LINK_STATE_UP_PACKET_RECEIVED)) { // パケット受信まだ.
			// パケットが来たかどうかのチェック.
			if (check_packet_receive(priv, false)) {
				// ハマり対策リセット.
				// TODO? XGMAC Reset.
				// Unpack Reset.
				ifrtp_reset_unpack(priv);

				phy_link_state |= LINK_STATE_UP_PACKET_RECEIVED; // 受信できたフラグを付与.
				pr_info("IF[%d] Up Packet Received\n", priv->netif_index);
			}
		}

		if (run) {
			if (phy_link_state & LINK_STATE_UP_PACKET_RECEIVED)
				open_rx_discard(priv->__netarch.dev);

			if (netif_queue_stopped(priv->__netarch.dev)) {
				netif_wake_queue(priv->__netarch.dev);
				pr_info("%s netif wake queuing\n", priv->__netarch.dev->name);
			}
		}
	}

	// 前回状態の更新.
	prev_phy_link_state[priv->netif_index] = phy_link_state;

}

static const struct net_device_ops igw_dev_ops = {
		.ndo_open = igw_open,
		.ndo_stop = igw_close,
		.ndo_get_stats64 = igw_get_stats64,
		.ndo_start_xmit = igw_send,
		.ndo_set_mac_address = igw_set_mac_address,
};

static const struct ethtool_ops igw_ethtool_ops = {
		.get_link_ksettings = igw_get_link_ksettings,
		.set_link_ksettings = igw_set_link_ksettings,
};

int igw_net_ether_setup(struct net_device *dev, const char *drv_name)
{
	struct igw_priv *priv;

	if (dev == NULL)
		return -EINVAL;

	ether_setup(dev);
	dev->netdev_ops = &igw_dev_ops;
	dev->watchdog_timeo = IGW_NET_LINKDETECTION_TIMER_TIMEOUT;
	netif_carrier_off(dev);
	snprintf(dev->name, sizeof(dev->name), drv_name);
	dev->ethtool_ops = &igw_ethtool_ops;

	priv = (struct igw_priv *)netdev_priv(dev);
	if (priv == NULL)
		return -EINVAL;

	int err = ifrtp_i2c_read_mac_addr(priv->netif_priority, dev->dev_addr);

	if (err < 0) {
		pr_err("Error Reading MAC Address: %d\n", err);
		return err;
	}
	if (!is_valid_ether_addr(dev->dev_addr))
		pr_err("Invalid MAC Address: %pM\n", dev->dev_addr);

	struct IocNetTxSourceMacAddressConfig mac_config;

	memcpy(mac_config.eth_src, dev->dev_addr, sizeof(mac_config.eth_src));
	if (priv->netif_priority)
		ifrtp_set_tx_source_secondary_mac_address(NULL, &mac_config);
	else
		ifrtp_set_tx_source_primary_mac_address(NULL, &mac_config);

	timer_setup(&priv->__netarch.link_check_timer, igw_timer_handler, 0);
	priv->__netarch.link_check_timer.expires = jiffies + IGW_NET_LINKDETECTION_TIMER_TIMEOUT;

	priv->__netarch.link_check_queue_work = kmalloc(sizeof(struct ifrtp_link_check_queue_work), GFP_KERNEL);
	if (priv->__netarch.link_check_queue_work == NULL)
		return -ENOMEM;

	priv->__netarch.link_check_queue_work->dev = priv;
	INIT_WORK((struct work_struct *)(priv->__netarch.link_check_queue_work), igw_link_check_handler);
	if (get_phy_link_state(priv, 0) == LINK_STATE_UP)
		netif_carrier_on(priv->__netarch.dev);

	return 0;
}

int igw_net_ether_release(struct net_device *dev)
{
	int ret = 0;
	struct igw_priv *priv;

	ret = igw_close(dev);

	priv = (struct igw_priv *)netdev_priv(dev);
	del_timer(&priv->__netarch.link_check_timer);

	flush_scheduled_work();
	kfree(priv->__netarch.link_check_queue_work);

	unregister_netdev(dev);
	return ret;
}


