// SPDX-License-Identifier: GPL-2.0
/*
 * Sony ifrtp SFP Ethernet device driver
 *
 * Copyright (c) 2023 Sony Corporation. All rights reserved.
 */
#include <linux/io.h>
#include <linux/iopoll.h>
#include <linux/ioport.h>
#include <linux/platform_device.h>
#include <linux/of_address.h>
#include <linux/of_device.h>
#include <linux/of_platform.h>
#include <linux/irqreturn.h>
#include <linux/io.h>

#include "igw_driver.h"
#include "io.h"
#include "ifrtp.h"
#include "ifrtp_driver.h"
#include "irq.h"
#include "ifrtp_mdio.h"
#include "ifrtp_address.h"
#include "igw_net_address.h"
#include "linux/if.h"

u8 __iomem *io_remapped, *led_io;

int ifrtp_led_on(unsigned int index)
{
	if (!IS_ERR(led_io))
		iowrite32(1, led_io + index * 4);
	else
		return PTR_ERR(led_io);

	return 0;
}

int ifrtp_led_off(unsigned int index)
{
	if (!IS_ERR(led_io))
		iowrite32(0, led_io + index * 4);
	else
		return PTR_ERR(led_io);

	return 0;
}

void ifrtp_irq_disable(struct igw_driver_handle *handle)
{
	ifrtp_disable_all_net_irq(handle->dev[0]);
}

void ifrtp_irq_clear(struct igw_driver_handle *handle)
{
}

bool ifrtp_irq_get_raw_packet(struct igw_priv *dev)
{
	return ifrtp_get_net_raw_packet_irq_status(dev);
}

void ifrtp_irq_clear_raw_packet(struct igw_priv *dev)
{
}

void ifrtp_irq_enable_raw_packet(struct igw_priv *dev)
{
	ifrtp_enable_net_raw_packet_irq(dev);
}

void ifrtp_irq_disable_raw_packet(struct igw_priv *dev)
{
	ifrtp_disable_net_raw_packet_irq(dev);
}

extern bool debugfs_link_sw_force_up;
bool ifrtp_sfp_is_linked_up(int if_no)
{
	if (debugfs_link_sw_force_up)
		return true;

	u32 lane_status = ifrtp_ior32(NULL, FPGA_MII_COMMON + IFRTP_OFFSET_SFP_LINKSTATE + if_no * IFRTP_OFFSET_SFP_LINKSTATE_CH);

	return (lane_status & IFRTP_SFP_STAT_LINK_OK_MASK) == IFRTP_SFP_STAT_LINK_OK;
}

u32 ifrtp_sfp_lane_status(int if_no)
{
	u32 lane_status = ifrtp_ior32(NULL, FPGA_MII_COMMON + IFRTP_OFFSET_SFP_LINKSTATE + if_no * IFRTP_OFFSET_SFP_LINKSTATE_CH);

	return lane_status;
}

int ifrtp_sfp_get_linkspeed(struct igw_priv *dev)
{
	return SPEED_1000;
}

int ifrtp_fpga_pll_reset(void)
{
	u32 retval = 0, res;
	unsigned int address = FPGA_MII_COMMON + IFRTP_OFFSET_PCS_PMA_RESET;

	ifrtp_iow32(NULL, 3, address);
	ifrtp_iow32(NULL, 0, address);

	// wait until pll reset done
	retval |= read_poll_timeout(ifrtp_ior32, res, (res & 3) == 0,
			1, 200000, false, NULL, address);
	retval |= read_poll_timeout(ifrtp_ior32, res, (res & 7) == 7,
			1, 200000, false, NULL, FPGA_MII_COMMON + IFRTP_OFFSET_SFP_LINKSTATE);
	retval |= read_poll_timeout(ifrtp_ior32, res, (res & 7) == 7,
			1, 200000, false, NULL, FPGA_MII_COMMON + IFRTP_OFFSET_SFP_LINKSTATE + IFRTP_OFFSET_SFP_LINKSTATE_CH);

	return retval;
}

int ifrtp_toggle_rx_datapath_reset(unsigned int if_mask)
{
	u32 retval = 0;
	unsigned int address = FPGA_MII_COMMON + IFRTP_OFFSET_RX_DATAPATH_RESET;

	retval = ifrtp_ior32(NULL, address);
	ifrtp_pr_debug("rx datapath reset 0x%08x, 0x%08x\n", retval, if_mask);
	ifrtp_iow32(NULL, retval ^ if_mask, address);

	return (retval & if_mask) ^ if_mask;
}
